rm(list=ls(all=TRUE))
library(compiler)

##########################################################################################################
# Inverse (gnralise) d'une matrice symmetrique semi-definie positive
Inv <- function(Sigma,tol=sqrt(.Machine$double.eps)) {
  n<- nrow(Sigma)
  sp<-eigen(Sigma,symmetric=TRUE)
  sp$vector%*%diag(1/pmax(abs(sp$values),tol))%*%t(sp$vector)
}
Inv <- cmpfun(Inv)

##########

Kro<- function(A,B){
A<-as.matrix(A)
B<-as.matrix(B)
lA<-dim(A)[1]
lB<-dim(B)[1]
cA<-dim(A)[2]
cB<-dim(B)[2]
res<-matrix(0,nrow=lA*lB,ncol=cA*cB)
for(i in 1:lA){
for(j in 1:cA){
res[((i-1)*lB+1):(i*lB),((j-1)*cB+1):(j*cB)]<-A[i,j]*B
}}
res
}

##############
# estimateur du paramtre DRM pour un GARCH(1,1)
##############
#
G<- function(xx,alpha.1=0.01,alpha.m=0.10,r=0.5){
sapply(xx, function(x){
res<-0
if(x>alpha.m)res<-1
if(x>alpha.1&x<=alpha.m)res<-((x-alpha.1)/(alpha.m-alpha.1))^r
res
})
}

##############
##Estimation des paramtres du modle GARCH (1,1) par un qmle bas sur la GED de paramtre kap 
#qu'on le dtermine en minimisant tau en utilisant
#les rsidus du GQMLE+ estimation des VaR parameters.
##############
##############
#######################################
# sigma.star estim lorsque h est une GED(kap)

sigma.star.GED<- function(kap, eta){
num<-kap*mean(abs(eta)^kap)
den<-2
sigma.star<-(num/den)^(1/kap)
sigma.star
 }
sigma.star.GED <- cmpfun(sigma.star.GED)

# tau estim lorsque h est une GED(kap)
obj.tau.GED<- function(kapp, eta){
sapply(kapp, function (kap){
num<-mean(abs(eta)^(2*kap))
den<-(mean(abs(eta)^(kap)))^2
tau<-4*(num/den-1)/(kap^2)
tau})
 }
obj.tau.GED <- cmpfun(obj.tau.GED)


# kap optimal dans cas ou h est une GED(kap)
kap.opt.GED<- function(eta){
kap.init<-2
res <- nlminb(kap.init,obj.tau.GED, lower=c(0.1),upper=c(5), 
   eta=eta)
kap.opt<-res$par
tau.opt<-res$objective
list(kap.opt=kap.opt,tau.opt=tau.opt)
 }
kap.opt.GED <- cmpfun(kap.opt.GED)


#######################################
# estime un GARCH(1,1) et le paramtre VaR  par QMLE non gaussien 
# fonde sur la fonction instrumentale GED
#

#         GED pdf
pdf.GED<- function (x,kap) {
densite<-exp(-0.5*abs(x)^(kap))/(gamma(1/kap+1)*2^(1/kap+1))
densite
}
pdf.GED <- cmpfun(pdf.GED)



# fonction objectif
objf.GED <- function(x, eps, kap, sig2.init, r0){     
         omega <- x[1]
         alpha <- x[2]
         beta <- x[3]
         n <- length(eps)
         sigma2<-rep(0,n)
         vect<-sigma2
sigma2[1]<-sig2.init
for (t in 2:n)sigma2[t]<-omega+alpha*eps[t-1]**2+beta*sigma2[t-1]

vect[2:n]<-pdf.GED(eps[2:n]/sqrt(sigma2[2:n]), kap)/sqrt(sigma2[2:n])
         qml <- -mean(log(vect[(r0+1):n]))
         qml }

objf.GED <- cmpfun(objf.GED)

## kap<-kap.opt; tau<-tau.opt.GED; sig<-sigma.star
##############
#omega.est<-omega1;alpha.est<-alpha1;beta.est<-beta1
#eps<-eps0;kap<-kap.opt;tau<-tau.opt.GED;sig<-sigma.star

VarAsympDRM.GED<- function(omega.est,alpha.est,beta.est,sig.init,eps,prev,kap,tau,m,
       vinit=10,alpha.1=0.01,alpha.m=0.10,r=0.5, grand=1/sqrt(.Machine$double.eps), alpha.0=0.05){
         n <- length(eps)
         sigma2<-rep(0,n)
         dersigma2<-matrix(0,nrow=3,ncol=n)
         Z<-dersigma2
sigma2[1]<-sig.init^2         
for (t in 2:n){
sigma2[t]<-omega.est+alpha.est*eps[t-1]^2+beta.est*sigma2[t-1]
dersigma2[1:3,t]<-c(1,eps[t-1]^2,sigma2[t-1])+beta.est*dersigma2[1:3,(t-1)]
Z[1:3,t]<-dersigma2[1:3,t]/sigma2[t]
}
J<-Z[1:3,vinit:n]%*%t(Z[1:3,vinit:n])/(n-vinit)
eta<-eps[vinit:n]/sqrt(sigma2[vinit:n]) #plot(ts(eta))
#eta<-eta/sd(eta)
eta.sort<-sort(eta)
T<-length(eta)

alpha.vect<-seq(alpha.1,alpha.m,length.out=m)

xi.n<-eta.sort[ceiling(T*alpha.vect)]

poids<-alpha.vect
poids[1]<-(alpha.vect[1]^r)/(alpha.vect[m]^r)
poids[2:m]<-sapply(2:m, function(i) (alpha.vect[i]^r-alpha.vect[i-1]^r)/(alpha.vect[m]^r))
#plot(poids) sum(poids)

DRM.eta<- -sum(poids*xi.n)

eta.kap<-abs(eta)^kap
g1<-1+eta.kap/mean(eta.kap)
g2<-1-(1+kap)*eta.kap/mean(eta.kap)
Eg2<-mean(g2)

p.alpha<-rep(0,m)
for(i in 1:m){
indicator<-sapply(1:T, function(t) ifelse(eta[t]<xi.n[i],1,0))
p.alpha[i]<-cov(indicator,g1)}

f<-density(eta)
f.xi<-sapply(xi.n,function(xi) f$y[which.min(abs(f$x-xi))])

zeta<-matrix(0,nrow=m,ncol=m)
for(i in 1:m){
for(j in 1:i){
zeta[i,j]<-{(xi.n[i]*xi.n[j]*tau/4 -xi.n[i]*p.alpha[j]/(Eg2*f.xi[j])-
xi.n[j]*p.alpha[i]/(Eg2*f.xi[i])+alpha.vect[j]*(1-alpha.vect[i])/(f.xi[i]*f.xi[j]))}
zeta[j,i]<-zeta[i,j]
}}

lambda.alphai<- (-xi.n*tau +4*p.alpha/(Eg2*f.xi))

theta0<-c(omega.est,alpha.est,0)
Jinv<-Inv(J)
Sigma<-rbind(cbind(Jinv*tau,Kro(t(lambda.alphai),theta0)/2),cbind(Kro(lambda.alphai,t(theta0)/2),zeta))

Delta<-matrix(0,nrow=3,ncol=(3+m))
Delta[1,1]<-DRM.eta^2
Delta[2,2]<-DRM.eta^2
Delta[3,3]<-1
Delta[1,4:(3+m)]<--2*omega.est*DRM.eta*poids
Delta[2,4:(3+m)]<--2*alpha.est*DRM.eta*poids
sdv<-sqrt(diag(Delta%*%Sigma%*%t(Delta)/T))


         n.prev <- length(prev)
         sigma2.prev<-rep(0,n.prev)
         DRM<-rep(0,n.prev)
         DRM.L<-DRM
         DRM.U<-DRM
cte<-qnorm(1-alpha.0/2)

sigma2.prev[1]<-sigma2[n]
dersigma2.prev<-matrix(0,nrow=3,ncol=n.prev)
Z.prev<-dersigma2.prev
Z.prev[1:3,1]<-Z[1:3,n]         
DRM[1]<- sqrt(sigma2[n])*DRM.eta
delta<-c(DRM.eta*dersigma2[1:3,n],sigma2[n]*poids)
sig<-sqrt(as.numeric(t(delta)%*%Sigma%*%delta/n))
DRM.L[1]<-DRM[1]-cte*sig
DRM.U[1]<-DRM[1]+cte*sig



for (t in 2:n.prev){
sigma2.prev[t]<-omega.est+alpha.est*prev[t-1]^2+beta.est*sigma2.prev[t-1]
dersigma2.prev[1:3,t]<-c(1,prev[t-1]^2,sigma2.prev[t-1])+beta.est*dersigma2.prev[1:3,(t-1)]
Z.prev[1:3,t]<-dersigma2.prev[1:3,t]/sigma2.prev[t]
DRM[t]<- sqrt(sigma2.prev[t])*DRM.eta
delta<-c(DRM.eta*dersigma2.prev[1:3,t],sigma2.prev[t]*poids)
sig<-sqrt(as.numeric(t(delta)%*%Sigma%*%delta/n))
DRM.L[t]<-DRM[t]-cte*sig
DRM.U[t]<-DRM[t]+cte*sig
}




list(coef.DRM=c(DRM.eta^2*omega.est,DRM.eta^2*alpha.est,beta.est), sdv.DRM=sdv, eta=eta,DRM=DRM,DRM.L=DRM.L,DRM.U=DRM.U)
}
VarAsympDRM.GED <- cmpfun(VarAsympDRM.GED)


#omega0<-omega.init;alpha0<-alpha.init;beta0<-beta.init;eps0<-ser.est;m<-5;prev<-ser.prev
#####################################################################################################
# Estime les paramtres DRM par QMLE GED
#####################################################################################################

estim.GED.DRM.garch11<- function(omega0,alpha0,beta0,eps0,eta,prev,m,vinit=10,alpha.1=0.01,alpha.m=0.10,r=0.5)
     {
n <- length(eps0)
sig.init<-0
{if(n>=5)sig.init<-sd(eps0[1:5])}



res<-kap.opt.GED(eta)
kap.opt<-res$kap.opt
tau.opt.GED<-res$tau.opt

sigma.star<-sigma.star.GED(kap.opt, eta)


valinit<-c(omega0*sigma.star^2,alpha0*sigma.star^2,beta0)
res <- nlminb(valinit,objf.GED, lower=c(0.0000001,0.0000001,0),upper=c(Inf,Inf,1), 
   eps=eps0,kap=kap.opt,sig2.init=sig.init^2,r0=vinit)


omega1<-res$par[1]
alpha1<-res$par[2]
beta1<-res$par[3]

resDRM<-VarAsympDRM.GED(omega1,alpha1,beta1,sig.init,eps0,prev,kap.opt,tau.opt.GED,m)

list(coef=c(omega1,alpha1,beta1),coef.DRM=resDRM$coef.DRM,sdv=resDRM$sdv.DRM,DRM=resDRM$DRM,DRM.L=resDRM$DRM.L,DRM.U=resDRM$DRM.U)
}
estim.GED.DRM.garch11 <- cmpfun(estim.GED.DRM.garch11)


#####################################################################################################
# Estime les paramtres DRM par QMLE gaussien
#####################################################################################################
##############
#omega.est<-omega1;alpha.est<-alpha1;beta.est<-beta1;eps<-eps0
VarAsympDRM<- function(omega.est,alpha.est,beta.est,sig.init,eps,prev,m,
       vinit=10,alpha.1=0.01,alpha.m=0.10,r=0.5, grand=1/sqrt(.Machine$double.eps), alpha.0=0.05){
         n <- length(eps)
         sigma2<-rep(0,n)
         dersigma2<-matrix(0,nrow=3,ncol=n)
         Z<-dersigma2
sigma2[1]<-sig.init^2         
for (t in 2:n){
sigma2[t]<-omega.est+alpha.est*eps[t-1]^2+beta.est*sigma2[t-1]
dersigma2[1:3,t]<-c(1,eps[t-1]^2,sigma2[t-1])+beta.est*dersigma2[1:3,(t-1)]
Z[1:3,t]<-dersigma2[1:3,t]/sigma2[t]
}
J<-Z[1:3,vinit:n]%*%t(Z[1:3,vinit:n])/(n-vinit)
J<-J/4
eta<-eps[vinit:n]/sqrt(sigma2[vinit:n]) #plot(ts(eta))
#eta<-eta/sd(eta)
eta.sort<-sort(eta)
T<-length(eta)

alpha.vect<-seq(alpha.1,alpha.m,length.out=m)

xi.n<-eta.sort[ceiling(T*alpha.vect)]

poids<-alpha.vect
poids[1]<-(alpha.vect[1]^r)/(alpha.vect[m]^r)
poids[2:m]<-sapply(2:m, function(i) (alpha.vect[i]^r-alpha.vect[i-1]^r)/(alpha.vect[m]^r))
#plot(poids) sum(poids)

DRM.eta<- -sum(poids*xi.n)


p.alpha<-sapply(1:m, function(i.alpha) {
indicator<-sapply(1:T, function(t) ifelse(eta[t]<xi.n[i.alpha],1,0))
res<-mean(indicator*eta^2)-alpha.vect[i.alpha]})

f<-density(eta)
f.xi<-sapply(xi.n,function(xi) f$y[which.min(abs(f$x-xi))])

kap4<-mean(eta^4)
zeta<-matrix(0,nrow=m,ncol=m)
for(i in 1:m){
for(j in 1:i){
zeta[i,j]<-{xi.n[i]*xi.n[j]*(kap4-1)/4 +xi.n[i]*p.alpha[j]/(2*f.xi[j])+
xi.n[j]*p.alpha[i]/(2*f.xi[i])+alpha.vect[j]*(1-alpha.vect[i])/(f.xi[i]*f.xi[j])}
zeta[j,i]<-zeta[i,j]
}}

lambda.alphai<-xi.n*(kap4-1)/4 +p.alpha/(2*f.xi)

theta0<-c(omega.est,alpha.est,0)
Jinv<-Inv(J)
Sigma<-rbind(cbind(Jinv*(kap4-1)/4,2*Kro(t(-lambda.alphai),theta0)),cbind(2*Kro(-lambda.alphai,t(theta0)),zeta))

Delta<-matrix(0,nrow=3,ncol=(3+m))
Delta[1,1]<-DRM.eta^2
Delta[2,2]<-DRM.eta^2
Delta[3,3]<-1
Delta[1,4:(3+m)]<--2*omega.est*DRM.eta*poids
Delta[2,4:(3+m)]<--2*alpha.est*DRM.eta*poids
sdv<-sqrt(diag(Delta%*%Sigma%*%t(Delta)/T))

         n.prev <- length(prev)
         sigma2.prev<-rep(0,n.prev)
         DRM<-rep(0,n.prev)
         DRM.L<-DRM
         DRM.U<-DRM
cte<-qnorm(1-alpha.0/2)

sigma2.prev[1]<-sigma2[n]
dersigma2.prev<-matrix(0,nrow=3,ncol=n.prev)
Z.prev<-dersigma2.prev
Z.prev[1:3,1]<-Z[1:3,n]         
DRM[1]<- sqrt(sigma2[n])*DRM.eta
delta<-c(DRM.eta*dersigma2[1:3,n],sigma2[n]*poids)
sig<-sqrt(as.numeric(t(delta)%*%Sigma%*%delta/n))
DRM.L[1]<-DRM[1]-cte*sig
DRM.U[1]<-DRM[1]+cte*sig



for (t in 2:n.prev){
sigma2.prev[t]<-omega.est+alpha.est*prev[t-1]^2+beta.est*sigma2.prev[t-1]
dersigma2.prev[1:3,t]<-c(1,prev[t-1]^2,sigma2.prev[t-1])+beta.est*dersigma2.prev[1:3,(t-1)]
Z.prev[1:3,t]<-dersigma2.prev[1:3,t]/sigma2.prev[t]
DRM[t]<- sqrt(sigma2.prev[t])*DRM.eta
delta<-c(DRM.eta*dersigma2.prev[1:3,t],sigma2.prev[t]*poids)
sig<-sqrt(as.numeric(t(delta)%*%Sigma%*%delta/n))
DRM.L[t]<-DRM[t]-cte*sig
DRM.U[t]<-DRM[t]+cte*sig
}




list(coef.DRM=c(DRM.eta^2*omega.est,DRM.eta^2*alpha.est,beta.est), sdv.DRM=sdv, eta=eta,DRM=DRM,DRM.L=DRM.L,DRM.U=DRM.U)


}
VarAsympDRM <- cmpfun(VarAsympDRM)

######

objf3 <- function(x, n, eps2, sig.init,vinit=10){     
         omega0 <- x[1]
         alpha0 <- x[2]
         beta0 <- x[3]
         sigma2<-rep(1,n)
         sigma2[1]<-sig.init^2
{for (t in 2:n)sigma2[t]<-omega0+alpha0*eps2[t-1]+beta0*sigma2[t-1]}
crit<-log(sigma2[vinit:n]) + eps2[vinit:n]/sigma2[vinit:n]
         qml <- mean(crit)
         qml }
objf3 <- cmpfun(objf3)

##################################
#omega0<-omega.init; alpha0<-alpha.init; beta0<-beta.init;eps0<-ser.est;prev<-ser.prev

estim.DRM.garch11<- function(omega0,alpha0,beta0,eps0,prev,m,vinit=10,alpha.1=0.01,alpha.m=0.10,r=0.5)
     {
n <- length(eps0)
sig.init<-0
{if(n>=5)sig.init<-sd(eps0[1:5])}
valinit<-c(omega0,alpha0,beta0)
res <- nlminb(valinit,objf3, lower=c(0.0000001,0,0),upper=c(Inf,Inf,1), 
 n=n,  eps2=eps0^2,sig.init=sig.init)



omega1<-res$par[1]
alpha1<-res$par[2]
beta1<-res$par[3]

resDRM<-VarAsympDRM(omega1,alpha1,beta1,sig.init,eps0,prev,m)

list(coef=c(omega1,alpha1,beta1),coef.DRM=resDRM$coef.DRM,sdv=resDRM$sdv.DRM,eta=resDRM$eta,DRM=resDRM$DRM,DRM.L=resDRM$DRM.L,DRM.U=resDRM$DRM.U)
}
estim.DRM.garch11 <- cmpfun(estim.DRM.garch11)

#############################################################################################


#####################################################
ptm <- Sys.time()
###########################################################################
############################################################################
###########################################################################
############################################################################
indices<-c("cac40.csv","dax.csv","ftse.csv","nikkei.csv",
 "nsei.csv", "smi.csv", "sp500.csv", "sptsx.csv", "sse.csv")
name.indices<-c("CAC","DAX","FTSE",
               "Nikkei","NSE", "SMI", "SP500","SPTSX","SSE")
#indices<-c("cac40.csv")
#name.indices<-c("CAC")
nbind<-length(indices)
grand=1/sqrt(.Machine$double.eps)
alpha.init<-0.05
beta.init<-0.85
omega.init<-0.01
ind<-1
m<-5

nb.prev<-100
data <- read.table(indices[ind],header=TRUE,sep=",")
close<-rev(data$Close) 
date<-rev(data$Date)
ntot<-length(close)
rend<-rep(0,ntot); rend[2:ntot]<-log(close[2:ntot]/close[1:(ntot-1)])*100 
ser0<-rend[2:ntot] 
n0<-length(ser0)
#date[length(ser0)]

#n0<-n0-100
################################################################
# 
nb.est<-n0-nb.prev
nb.est<-1000
ser.est<-ser0[(n0-nb.est-nb.prev+1):(n0-nb.prev)]
date.est<-date[(n0-nb.est-nb.prev+2):(n0-nb.prev+1)]
ser.prev<-ser0[(n0-nb.prev+1):n0]
date.prev<-date[(n0-nb.prev+2):(n0+1)]

# DRM par QML Gaussien

res<-estim.DRM.garch11(omega.init,alpha.init,beta.init,ser.est,ser.prev,m)
omega.est<-res$coef.DRM[1]
alpha.est<-res$coef.DRM[2]
beta.est<-res$coef.DRM[3]

#
DRM.est<-res$DRM
DRM.moins<-res$DRM.L
DRM.plus<-res$DRM.U


# DRM par QML GED
eta<-res$eta

res.GED<-estim.GED.DRM.garch11(omega.init,alpha.init,beta.init,ser.est,eta,ser.prev,m)
omega.est.GED<-res.GED$coef.DRM[1]
alpha.est.GED<-res.GED$coef.DRM[2]
beta.est.GED<-res.GED$coef.DRM[3]

#
DRM.est.GED<-res.GED$DRM
DRM.moins.GED<-res.GED$DRM.L
DRM.plus.GED<-res.GED$DRM.U




##########################################
min<-min(ser.prev,-DRM.plus.GED,-DRM.plus)
max<-max(ser.prev)

plot(ts(ser.prev),axes=FALSE,ylim=c(min,max),ylab='Return and -DRM',
main=paste(name.indices[ind], ': DRM and DRM accuracy intervals estimated by two methods'),xlab=NULL,col="blue",cex.main=0.9)
lines(-DRM.est,lty=2,lwd=2,col="red")
lines(-DRM.moins,lty=2,lwd=1,col="red")
lines(-DRM.plus,lty=2,lwd=1,col="red")

lines(-DRM.est.GED,lwd=2,col="seagreen")
lines(-DRM.moins.GED,lty=1,lwd=1,col="seagreen")
lines(-DRM.plus.GED,lty=1,lwd=1,col="seagreen")

box()
axis(2)
Ordrerecord.moins<-order(ser.prev)[1:1]
Rendement.moins<-ser.prev[Ordrerecord.moins]
Date.rendement.moins<-date.prev[Ordrerecord.moins]
lab1<-c("",as.character(Date.rendement.moins),"")
at1<-c(1,Ordrerecord.moins,nb.prev)
axis(1,at=at1, lab=lab1, cex.axis=0.9)



